@@ -26,6 +26,7 @@ from progress.spinner import Spinner
 from .scanossapi import ScanossApi
 from .winnowing import Winnowing
 from .cyclonedx import CycloneDx
+from .threadedscanning import ThreadedScanning
 
 FILTERED_DIRS = {  # Folders to skip
                  "nbproject", "nbbuild", "nbdist", "__pycache__", "venv", "_yardoc", "eggs", "wheels", "htmlcov",
@@ -67,14 +68,14 @@ MAX_POST_SIZE = 64 * 1024  # 64k Max post size
 
 class Scanner:
     """
-
+    SCANOSS scanning class
     """
     def __init__(self, wfp: str = None, scan_output: str = None, output_format: str = 'plain',
                  debug: bool = False, trace: bool = False, quiet: bool = False, api_key: str = None, url: str = None,
-                 sbom_path: str = None, scan_type: str = None, flags: str = None
+                 sbom_path: str = None, scan_type: str = None, flags: str = None, nb_threads: int = 5
                  ):
         """
-
+        Initialise scanning class, including Winnowing, ScanossApi and ThreadedScanning
         """
         self.quiet = quiet
         self.debug = debug
@@ -87,6 +88,13 @@ class Scanner:
         self.scanoss_api = ScanossApi(debug=debug, trace=trace, quiet=quiet, api_key=api_key, url=url,
                                       sbom_path=sbom_path, scan_type=scan_type, flags=flags
                                       )
+        self.nb_threads = nb_threads
+        if nb_threads and nb_threads > 0:
+            self.threaded_scan = ThreadedScanning(self.scanoss_api, debug=debug, trace=trace, quiet=quiet,
+                                                  nb_threads=nb_threads
+                                                  )
+        else:
+            self.threaded_scan = None
 
     @staticmethod
     def __filter_files(files) -> list:
@@ -179,7 +187,6 @@ class Scanner:
                         count += 1
         return count
 
-
     @staticmethod
     def valid_json_file(json_file: str) -> bool:
         """
@@ -201,7 +208,6 @@ class Scanner:
             return False
         return True
 
-
     def print_msg(self, *args, **kwargs):
         """
         Print message if quite mode is not enabled
@@ -235,10 +241,15 @@ class Scanner:
         else:
             print(string)
 
-    def scan_folder(self, scan_dir: str):
+    def scan_folder(self, scan_dir: str) -> bool:
         """
         Scan the specified folder producing fingerprints, send to the SCANOSS API and return results
+
+        :param scan_dir: str
+                    Directory to scan
+        :return True if successful, False otherwise
         """
+        success = True
         if not scan_dir:
             raise Exception(f"ERROR: Please specify a folder to scan")
         if not os.path.exists(scan_dir) or not os.path.isdir(scan_dir):
@@ -246,6 +257,7 @@ class Scanner:
         wfps = ''
         scan_dir_len = len(scan_dir) if scan_dir.endswith(os.path.sep) else len(scan_dir)+1
         self.print_msg(f'Searching {scan_dir} for files to fingerprint...')
+        spinner = None
         if not self.quiet and self.isatty:
             spinner = Spinner('Fingerprinting ')
         for root, dirs, files in os.walk(scan_dir):
@@ -258,10 +270,10 @@ class Scanner:
                 file_stat = os.stat(path)
                 if file_stat.st_size > 0:            # Ignore empty files
                     self.print_debug(f'Fingerprinting {path}...')
-                    if not self.quiet and self.isatty:
+                    if spinner:
                         spinner.next()
                     wfps += self.winnowing.wfp_for_file(path, Scanner.__strip_dir(scan_dir, scan_dir_len, path))
-        if not self.quiet and self.isatty:
+        if spinner:
             spinner.finish()
         if wfps:
             self.print_debug(f'Writing fingerprints to {self.wfp}')
@@ -270,18 +282,24 @@ class Scanner:
             self.print_msg(f'Scanning fingerprints...')
             if self.scan_output:
                 self.print_msg(f'Writing results to {self.scan_output}...')
-            self.scan_wfp_file()
+            if self.threaded_scan:
+                success = self.scan_wfp_file_threaded()
+            else:
+                success = self.scan_wfp_file()
         else:
             Scanner.print_stderr(f'Warning: No files found to scan in folder: {scan_dir}')
+        return success
 
-    def scan_file(self, file: str):
+    def scan_file(self, file: str) -> bool:
         """
         Scan the specified file and produce a result
         Parameters
         ----------
             file: str
                 File to fingerprint and scan/identify
+        :return True if successful, False otherwise
         """
+        success = True
         if not file:
             raise Exception(f"ERROR: Please specify a file to scan")
         if not os.path.exists(file) or not os.path.isfile(file):
@@ -292,16 +310,113 @@ class Scanner:
             self.print_debug(f'Scanning {file}...')
             if self.scan_output:
                 self.print_msg(f'Writing results to {self.scan_output}...')
-            self.scan_wfp(wfps)
+            success = self.scan_wfp(wfps)
+        else:
+            success = False
 
-    def scan_wfp_file(self, file: str = None):
+        return success
+
+    def scan_wfp_file_threaded(self, file: str = None) -> bool:
         """
-        Scan the contents of the specified WFP file
+        Scan the supplied WFP file in multiple threads
+        :param file: WFP file to scan
+        :return True if scuccessful, False otherwise
+        """
+        success = True
+        wfp_file = file if file else self.wfp   # If a WFP file is specified, use it, otherwise us the default
+        if not os.path.exists(wfp_file) or not os.path.isfile(wfp_file):
+            raise Exception(f"ERROR: Specified WFP file does not exist or is not a file: {wfp_file}")
+        file_count = Scanner.__count_files_in_wfp_file(wfp_file)
+        cur_files = 0
+        cur_size = 0
+        batch_files = 0
+        wfp = ''
+        self.print_debug(f'Found {file_count} files to process.')
+        file_print = ''
+        bar = None
+        if not self.quiet and self.isatty:
+            bar = Bar('Scanning', max=file_count)
+            bar.next(0)
+            self.threaded_scan.set_bar(bar)
+
+        with open(wfp_file) as f:
+            for line in f:
+                if line.startswith(WFP_FILE_START):
+                    if file_print:
+                        wfp += file_print         # Store the WFP for the current file
+                        cur_size = len(wfp.encode("utf-8"))
+                    file_print = line             # Start storing the next file
+                    cur_files += 1
+                    batch_files += 1
+                else:
+                    file_print += line             # Store the rest of the WFP for this file
+                l_size = cur_size + len(file_print.encode('utf-8'))
+                # Hit the max post size, so sending the current batch and continue processing
+                if l_size >= MAX_POST_SIZE and wfp:
+                    self.print_debug(f'Added {batch_files} ({cur_files}) of'
+                                     f' {file_count} ({len(wfp.encode("utf-8"))} bytes) files to the pending queue.')
+                    if cur_size > MAX_POST_SIZE:
+                        Scanner.print_stderr(f'Warning: Post size {cur_size} greater than limit {MAX_POST_SIZE}')
+                    self.threaded_scan.queue_add(wfp)
+                    batch_files = 0
+                    wfp = ''
+        if file_print:
+            wfp += file_print  # Store the WFP for the current file
+        if wfp:
+            self.print_debug(f'Adding {batch_files} ({cur_files}) of'
+                             f' {file_count} ({len(wfp.encode("utf-8"))} bytes) files to the pending queue.')
+            self.threaded_scan.queue_add(wfp)
+
+        if not self.threaded_scan.run():
+            self.print_stderr(f'Warning: Some errors encounted while scanning. Result might not be complete.')
+            success = False
+        responses = self.threaded_scan.responses
+        raw_output = "{\n"
+        if responses:
+            first = True
+            for scan_resp in responses:
+                if scan_resp is not None:
+                    for key, value in scan_resp.items():
+                        if first:
+                            raw_output += "  \"%s\":%s" % (key, json.dumps(value, indent=2))
+                            first = False
+                        else:
+                            raw_output += ",\n  \"%s\":%s" % (key, json.dumps(value, indent=2))
+        else:
+            success = False
+        raw_output += "\n}"
+        if bar:
+            bar.finish()
+        parsed_json = None
+        try:
+            parsed_json = json.loads(raw_output)
+        except Exception as e:
+            self.print_stderr(f'Warning: Problem decoding parsed json: {e}')
+
+        if self.output_format == 'plain':
+            if parsed_json:
+                self.__log_result(json.dumps(parsed_json, indent=2, sort_keys=True))
+            else:
+                self.__log_result(raw_output)
+        elif self.output_format == 'cyclonedx':
+            cdx = CycloneDx(self.debug, self.scan_output)
+            if parsed_json:
+                success = cdx.produce_from_json(parsed_json)
+            else:
+                success = cdx.produce_from_str(raw_output)
+
+        return success
+
+    def scan_wfp_file(self, file: str = None) -> bool:
+        """
+        Scan the contents of the specified WFP file (in the current process)
         Parameters
         ----------
             file: str
                 WFP file to scan (optional)
+        return: True if successful, False otherwise
         """
+        success = True
         wfp_file = file if file else self.wfp   # If a WFP file is specified, use it, otherwise us the default
         if not os.path.exists(wfp_file) or not os.path.isfile(wfp_file):
             raise Exception(f"ERROR: Specified WFP file does not exist or is not a file: {wfp_file}")
@@ -315,6 +430,7 @@ class Scanner:
         self.print_debug(f'Found {file_count} files to process.')
         raw_output = "{\n"
         file_print = ''
+        bar = None
         if not self.quiet and self.isatty:
             bar = Bar('Scanning', max=file_count)
             bar.next(0)
@@ -337,7 +453,7 @@ class Scanner:
                     if cur_size > MAX_POST_SIZE:
                         Scanner.print_stderr(f'Warning: Post size {cur_size} greater than limit {MAX_POST_SIZE}')
                     scan_resp = self.scanoss_api.scan(wfp, max_component['name'])  # Scan current WFP and store
-                    if not self.quiet and self.isatty:
+                    if bar:
                         bar.next(batch_files)
                     if scan_resp is not None:
                         for key, value in scan_resp.items():
@@ -352,6 +468,8 @@ class Scanner:
                                             max_component['hits'] = components[vcv]
                                 else:
                                     Scanner.print_stderr(f'Warning: Unknown value: {v}')
+                    else:
+                        success = False
                     batch_files = 0
                     wfp = ''
         if file_print:
@@ -360,7 +478,7 @@ class Scanner:
             self.print_debug(f'Sending {batch_files} ({cur_files}) of'
                              f' {file_count} ({len(wfp.encode("utf-8"))} bytes) files to the ScanOSS API.')
             scan_resp = self.scanoss_api.scan(wfp, max_component['name'])  # Scan current WFP and store
-            if not self.quiet and self.isatty:
+            if bar:
                 bar.next(batch_files)
             first = True
             if scan_resp is not None:
@@ -370,8 +488,10 @@ class Scanner:
                         first = False
                     else:
                         raw_output += ",\n  \"%s\":%s" % (key, json.dumps(value, indent=2))
+            else:
+                success = False
         raw_output += "\n}"
-        if not self.quiet and self.isatty:
+        if bar:
             bar.finish()
         if self.output_format == 'plain':
             self.__log_result(raw_output)
@@ -379,7 +499,9 @@ class Scanner:
             cdx = CycloneDx(self.debug, self.scan_output)
             cdx.produce_from_str(raw_output)
 
-    def scan_wfp(self, wfp: str):
+        return success
+
+    def scan_wfp(self, wfp: str) -> bool:
         """
         Send the specified (single) WFP to ScanOSS for identification
         Parameters
@@ -387,6 +509,7 @@ class Scanner:
             wfp: str
                 Winnowing Fingerprint to scan/identify
         """
+        success = True
         if not wfp:
             raise Exception(f"ERROR: Please specify a WFP to scan")
         raw_output = "{\n"
@@ -394,6 +517,8 @@ class Scanner:
         if scan_resp is not None:
             for key, value in scan_resp.items():
                 raw_output += "  \"%s\":%s" % (key, json.dumps(value, indent=2))
+        else:
+            success = False
         raw_output += "\n}"
         if self.output_format == 'plain':
             self.__log_result(raw_output)
@@ -401,6 +526,8 @@ class Scanner:
             cdx = CycloneDx(self.debug, self.scan_output)
             cdx.produce_from_str(raw_output)
 
+        return success
+
     def wfp_file(self, scan_file: str, wfp_file: str = None):
         """
         Fingerprint the specified file
